from typing import Dict

KEYS: Dict[str, str] = {
    "`": "grave",
    "-": "minus",
    "=": "equal",
    "[": "bracketleft",
    "]": "bracketright",
    "\\": "backslash",
    ";": "semicolon",
    "'": "apostrophe",
    ",": "comma",
    ".": "period",
    "/": "slash",
    "bs": "BackSpace",
    "BS": "BackSpace",
    "Space": "space",
    "left": "Left",
    "right": "Right",
    "up": "Up",
    "down": "Down",
    "tab": "Tab",
    "f1": "F1",
    "f2": "F2",
    "f3": "F3",
    "f4": "F4",
    "f5": "F5",
    "f6": "F6",
    "f7": "F7",
    "f8": "F8",
    "f9": "F9",
    "f10": "F10",
    "f11": "F11",
    "f12": "F12",
    "insert": "Insert",
    "ins": "Insert",
    "delete": "Delete",
    "del": "Delete",
    "home": "Home",
    "end": "End",
    "pgup": "Prior",
    "PgUp": "Prior",
    "PageUp": "Prior",
    "pgdown": "Next",
    "PgDown": "Next",
    "PageDown": "Next",
    "menu": "Menu",
    "enter": "Return",
    "Enter": "Return",
    "return": "Return",
}
