from . import wm as wm, wrappers as wrappers
from typing import Optional

class Actions(wm.Actions):
    def __init__(self, window_manager: wm.WM) -> None: ...
    def focus_next(self, window: Optional[wrappers.Window] = ...) -> None: ...
    def focus_prev(self, window: Optional[wrappers.Window] = ...) -> None: ...
    def restart(self) -> None: ...
    def activate_window_desktop(self, window: wrappers.Window) -> Optional[bool]: ...
